@@ -2,7 +2,6 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2005-2008  Marcel Holtmann <marcel@holtmann.org>
  *  Copyright (C) 2006-2010  Bastien Nocera <hadess@hadess.net>
  *
  *
@@ -32,6 +31,7 @@
 #include "cc-bluetooth-panel.h"
 
 #include <bluetooth-client.h>
+#include <bluetooth-utils.h>
 #include <bluetooth-client-private.h>
 #include <bluetooth-killswitch.h>
 #include <bluetooth-chooser.h>
@@ -208,6 +208,47 @@ set_notebook_page (CcBluetoothPanel *self,
 	gtk_notebook_set_current_page (GTK_NOTEBOOK (WID ("props_notebook")), page);
 }
 
+static void
+add_extra_setup_widgets (CcBluetoothPanel *self,
+			 const char       *bdaddr)
+{
+	GValue value = { 0 };
+	char **uuids;
+	GList *list, *l;
+	GtkWidget *container;
+
+	if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (self->priv->chooser),
+							"uuids", &value) == FALSE)
+		return;
+
+	uuids = (char **) g_value_get_boxed (&value);
+	list = bluetooth_plugin_manager_get_widgets (bdaddr, (const char **) uuids);
+	if (list == NULL) {
+		g_value_unset (&value);
+		return;
+	}
+
+	container = WID ("additional_setup_box");
+	for (l = list; l != NULL; l = l->next) {
+		GtkWidget *widget = l->data;
+		gtk_box_pack_start (GTK_BOX (container), widget,
+				    FALSE, FALSE, 0);
+		gtk_widget_show_all (widget);
+	}
+	gtk_widget_show (container);
+	g_value_unset (&value);
+}
+
+static void
+remove_extra_setup_widgets (CcBluetoothPanel *self)
+{
+	GtkWidget *box;
+
+	box = WID ("additional_setup_box");
+	gtk_container_forall (GTK_CONTAINER (box), (GtkCallback) gtk_widget_destroy, NULL);
+	gtk_widget_hide (WID ("additional_setup_box"));
+}
+
 static void
 cc_bluetooth_panel_update_properties (CcBluetoothPanel *self)
 {
@@ -218,17 +259,19 @@ cc_bluetooth_panel_update_properties (CcBluetoothPanel *self)
 	g_signal_handlers_block_by_func (button, switch_connected_active_changed, self);
 
 	/* Hide all the buttons now, and show them again if we need to */
-	gtk_widget_hide (WID ("keyboard_button"));
-	gtk_widget_hide (WID ("sound_button"));
-	gtk_widget_hide (WID ("mouse_button"));
+	gtk_widget_hide (WID ("keyboard_box"));
+	gtk_widget_hide (WID ("sound_box"));
+	gtk_widget_hide (WID ("mouse_box"));
+	gtk_widget_hide (WID ("browse_box"));
+	gtk_widget_hide (WID ("send_box"));
+
+	/* Remove the extra setup widgets */
+	remove_extra_setup_widgets (self);
 
 	bdaddr = bluetooth_chooser_get_selected_device (BLUETOOTH_CHOOSER (self->priv->chooser));
 	if (bdaddr == NULL) {
 		gtk_widget_set_sensitive (WID ("properties_vbox"), FALSE);
 		gtk_switch_set_active (button, FALSE);
-		gtk_label_set_text (GTK_LABEL (WID ("paired_label")), "");
-		gtk_label_set_text (GTK_LABEL (WID ("type_label")), "");
-		gtk_label_set_text (GTK_LABEL (WID ("address_label")), "");
 		gtk_widget_set_sensitive (WID ("button_delete"), FALSE);
 		set_notebook_page (self, NOTEBOOK_PAGE_EMPTY);
 	} else {
@@ -259,26 +302,45 @@ cc_bluetooth_panel_update_properties (CcBluetoothPanel *self)
 		gtk_widget_set_sensitive (GTK_WIDGET (button), (services != NULL));
 		g_value_unset (&value);
 
+		/* UUIDs */
+		if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (self->priv->chooser),
+								"uuids", &value)) {
+			const char **uuids;
+			guint i;
+
+			uuids = (const char **) g_value_get_boxed (&value);
+			for (i = 0; uuids[i] != NULL; i++) {
+				if (g_str_equal (uuids[i], "OBEXObjectPush"))
+					gtk_widget_show (WID ("send_box"));
+				else if (g_str_equal (uuids[i], "OBEXFileTransfer"))
+					gtk_widget_show (WID ("browse_box"));
+			}
+			g_value_unset (&value);
+		}
+
 		/* Type */
 		type = bluetooth_chooser_get_selected_device_type (BLUETOOTH_CHOOSER (self->priv->chooser));
 		gtk_label_set_text (GTK_LABEL (WID ("type_label")), bluetooth_type_to_string (type));
 		switch (type) {
 		case BLUETOOTH_TYPE_KEYBOARD:
-			gtk_widget_show (WID ("keyboard_button"));
+			gtk_widget_show (WID ("keyboard_box"));
 			break;
 		case BLUETOOTH_TYPE_MOUSE:
 		case BLUETOOTH_TYPE_TABLET:
-			gtk_widget_show (WID ("mouse_button"));
+			gtk_widget_show (WID ("mouse_box"));
 			break;
 		case BLUETOOTH_TYPE_HEADSET:
 		case BLUETOOTH_TYPE_HEADPHONES:
 		case BLUETOOTH_TYPE_OTHER_AUDIO:
-			gtk_widget_show (WID ("sound_button"));
+			gtk_widget_show (WID ("sound_box"));
 		default:
-			/* FIXME others? */
+			/* others? */
 			;
 		}
 
+		/* Extra widgets */
+		add_extra_setup_widgets (self, bdaddr);
+
 		gtk_label_set_text (GTK_LABEL (WID ("address_label")), bdaddr);
 		g_free (bdaddr);
 
@@ -359,7 +421,8 @@ cc_bluetooth_panel_update_power (CcBluetoothPanel *self)
 	}
 
 	g_free (bdaddr);
-	gtk_widget_set_sensitive (WID ("hbox2") , sensitive);
+	gtk_widget_set_sensitive (WID ("box_power") , sensitive);
+	gtk_widget_set_sensitive (WID ("box_vis") , sensitive);
 
 	button = GTK_SWITCH (WID ("switch_bluetooth"));
 
@@ -372,21 +435,64 @@ static void
 keyboard_callback (GtkButton        *button,
 		   CcBluetoothPanel *self)
 {
-	launch_command(KEYBOARD_PREFS);
+	launch_command (KEYBOARD_PREFS);
 }
 
 static void
 mouse_callback (GtkButton        *button,
 		CcBluetoothPanel *self)
 {
-	launch_command(MOUSE_PREFS);
+	launch_command (MOUSE_PREFS);
 }
 
 static void
 sound_callback (GtkButton        *button,
 		CcBluetoothPanel *self)
 {
-	launch_command(SOUND_PREFS);
+	launch_command (SOUND_PREFS);
+}
+
+static void
+send_callback (GtkButton        *button,
+	       CcBluetoothPanel *self)
+{
+	char *bdaddr, *alias;
+
+	bdaddr = bluetooth_chooser_get_selected_device (BLUETOOTH_CHOOSER (self->priv->chooser));
+	alias = bluetooth_chooser_get_selected_device_name (BLUETOOTH_CHOOSER (self->priv->chooser));
+
+	bluetooth_send_to_address (bdaddr, alias);
+
+	g_free (bdaddr);
+	g_free (alias);
+}
+
+static void
+mount_finish_cb (GObject *source_object,
+		 GAsyncResult *res,
+		 gpointer user_data)
+{
+	GError *error = NULL;
+
+	if (bluetooth_browse_address_finish (source_object, res, &error) == FALSE) {
+		g_printerr ("Failed to mount OBEX volume: %s", error->message);
+		g_error_free (error);
+		return;
+	}
+}
+
+static void
+browse_callback (GtkButton        *button,
+		 CcBluetoothPanel *self)
+{
+	char *bdaddr;
+
+	bdaddr = bluetooth_chooser_get_selected_device (BLUETOOTH_CHOOSER (self->priv->chooser));
+
+	bluetooth_browse_address (G_OBJECT (self), bdaddr,
+				  GDK_CURRENT_TIME, mount_finish_cb, NULL);
+
+	g_free (bdaddr);
 }
 
 /* Visibility/Discoverable */
@@ -553,7 +659,7 @@ cc_bluetooth_panel_init (CcBluetoothPanel *self)
 	/* Note that this will only ever show the devices on the default
 	 * adapter, this is on purpose */
 	self->priv->chooser = bluetooth_chooser_new (NULL);
-	gtk_box_pack_start (GTK_BOX (WID ("box1")), self->priv->chooser, TRUE, TRUE, 0);
+	gtk_box_pack_start (GTK_BOX (WID ("box_devices")), self->priv->chooser, TRUE, TRUE, 0);
 	g_object_set (self->priv->chooser,
 		      "show-searching", FALSE,
 		      "show-device-type", FALSE,
@@ -566,11 +672,13 @@ cc_bluetooth_panel_init (CcBluetoothPanel *self)
 	column = bluetooth_chooser_get_type_column (BLUETOOTH_CHOOSER (self->priv->chooser));
 	gtk_tree_view_column_set_visible (column, FALSE);
 	widget = bluetooth_chooser_get_treeview (BLUETOOTH_CHOOSER (self->priv->chooser));
-	g_object_set (G_OBJECT (widget), "headers-visible", FALSE, NULL);
+	column = bluetooth_chooser_get_device_column (BLUETOOTH_CHOOSER (self->priv->chooser));
+	gtk_tree_view_column_set_title (column, _("Devices"));
 
 	/* Join treeview and buttons */
 	widget = bluetooth_chooser_get_scrolled_window (BLUETOOTH_CHOOSER (self->priv->chooser));
 	gtk_scrolled_window_set_min_content_height (GTK_SCROLLED_WINDOW (widget), 250);
+	gtk_scrolled_window_set_min_content_width (GTK_SCROLLED_WINDOW (widget), 200);
 	context = gtk_widget_get_style_context (widget);
 	gtk_style_context_set_junction_sides (context, GTK_JUNCTION_BOTTOM);
 	widget = WID ("toolbar");
@@ -592,6 +700,10 @@ cc_bluetooth_panel_init (CcBluetoothPanel *self)
 			  G_CALLBACK (keyboard_callback), self);
 	g_signal_connect (G_OBJECT (WID ("sound_button")), "clicked",
 			  G_CALLBACK (sound_callback), self);
+	g_signal_connect (G_OBJECT (WID ("browse_button")), "clicked",
+			  G_CALLBACK (browse_callback), self);
+	g_signal_connect (G_OBJECT (WID ("send_button")), "clicked",
+			  G_CALLBACK (send_callback), self);
 	g_signal_connect (G_OBJECT (WID ("switch_connection")), "notify::active",
 			  G_CALLBACK (switch_connected_active_changed), self);
 
